
NAME = 'pansy'

CMD_CLIENT_REQ              = 10  # 透传client请求
CMD_CLIENT_CLOSED           = 20  # 客户端连接被关闭

CMD_WRITE_TO_WORKER         = 100 # 给worker发送消息

CMD_WORKER_ASK_FOR_TASK     = 210 # 请求任务

CMD_WRITE_TO_USERS          = 230 # 主动下发
CMD_CLOSE_USERS             = 250 # 关闭多个客户端


DEFAULT_CONFIG = {
    'HOST': '127.0.0.1',
    'PORT': 9700,

    # 启动的room id 列表，闭区间
    'ROOM_ID_BEGIN': None,
    'ROOM_ID_END': None,
    # 同时并发房间数
    'ROOM_CONCURRENT': 1,

    'DEBUG': False,

    'BOX_CLASS': 'netkit.box.Box',

    # master class
    'MASTER_CLASS': 'pansy.master.Master',

    # station class
    'STATION_CLASS': 'pansy.station.Station',

    # worker class
    'WORKER_CLASS': 'pansy.worker.Worker',

    'NAME': NAME,

    'CONN_TIMEOUT': 3,

    # 重连等待时间。worker和client都在用
    'RECONNECT_INTERVAL': 1,

    'STOP_TIMEOUT': None,

    # 处理task超时(秒). 超过后会打印fatal日志. None 代表永不超时
    'WORK_TIMEOUT': None,

    # 等待处理的tasks的最大值。<=0代表无限
    'TASKS_MAX_SIZE': -1,
}




import errno
import socket
import time
import functools
from threading import Lock
from ..share.log import logger
from ..share import constants


def lock_write(func):
    @functools.wraps(func)
    def func_wrapper(client, *args, **kwargs):
        try:
            client.acquire_write_lock()
            return func(client, *args, **kwargs)
        finally:
            client.release_write_lock()
    return func_wrapper


class Client(object):

    address = None
    ensure = None
    sock = None

    # 加锁来保证线程安全
    write_lock = None

    def __init__(self, address, ensure, lock=True):
        self.address = address
        self.ensure = ensure

        if lock:
            self.write_lock = Lock()

    @lock_write
    def write(self, data):
        """
        发送
        :param data:
        :return:
        """

        while True:
            self._detect_connection_status()

            # 先要保证连接可用
            if self.closed() and not self._connect():
                logger.error('connect fail, address: %s%s',
                             self.address,
                             ', sleep %ss' % constants.DEFAULT_CONFIG['RECONNECT_INTERVAL'] if self.ensure else '')
                if self.ensure:
                    time.sleep(constants.DEFAULT_CONFIG['RECONNECT_INTERVAL'])
                    continue
                else:
                    return False

            if self._write_to_fd(data):
                # 写入成功
                return True
            else:
                if not self.ensure:
                    return False

    def close(self):
        """
        关闭连接
        :return:
        """
        if self.sock:
            try:
                self.sock.close()
            except:
                pass

            self.sock = None

    def closed(self):
        """
        检查连接
        :return:
        """

        return not self.sock

    def _detect_connection_status(self):
        """
        通过recv的方式，检查连接状态
        :return: 是否连接
        """

        if not self.sock:
            return False

        try:
            # setblocking也是可能抛异常的
            # File "/usr/local/lib/python2.7/socket.py", line 224, in meth
            #     return getattr(self._sock,name)(*args)
            # File "/usr/local/lib/python2.7/socket.py", line 170, in _dummy
            #     raise error(EBADF, 'Bad file descriptor')
            # error: [Errno 9] Bad file descriptor
            self.sock.setblocking(False)

            if not self.sock.recv(1):
                # 返回''，代表连接已经关闭
                self.close()
                return False
            else:
                # 收到了数据，连接正常
                # 理论上不会到达这里，因为trigger是不会收到回应的
                return True
        except socket.error as e:
            if e.errno in (errno.EINTR, errno.EAGAIN):
                # 中断/没有可读数据，连接正常
                return True
            # 其他情况下，都代表错误
            self.close()
            return False
        except KeyboardInterrupt:
            raise
        except:
            self.close()
            return False
        finally:
            if self.sock:
                self.sock.setblocking(True)

    def _write_to_fd(self, data):
        """
        写入
        """

        while data:
            try:
                num_bytes = self.sock.send(data)
            except KeyboardInterrupt:
                raise
            except Exception as e:
                logger.error('socket send fail. e: %s, data: %r', e, data[:100])
                # 关闭连接
                self.close()
                return False

            data = data[num_bytes:]

        return True

    def _connect(self):
        """
        在写入之前调用，保证连接可用
        :return:
        """
        if ':' in self.address[0]:
            # IPV6
            socket_type = socket.AF_INET6
        else:
            socket_type = socket.AF_INET

        self.sock = socket.socket(socket_type, socket.SOCK_STREAM)
        # TCP_NODELAY 打开
        self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        try:
            self.sock.connect(self.address)
        except KeyboardInterrupt:
            self.close()
            raise
        except:
            self.close()
            return False
        else:
            # 连接成功
            return True

    def acquire_write_lock(self):
        if self.write_lock:
            self.write_lock.acquire()

    def release_write_lock(self):
        if self.write_lock:
            self.write_lock.release()




import base64
import random

from .easy_aes import EasyAES
from ...share.log import logger


class TokenHelper:

    secret = None
    magic = None

    # 分割符
    split_bytes = b','

    block_size = 16

    def __init__(self, secret, magic):
        """
        secret: AES key,  must be either 16, 24, or 32 bytes long
        magic: 作用是为了检测解密是否正确
        """
        self.secret = self._force_convert_to_bytes(secret)
        self.magic = self._force_convert_to_bytes(magic)

    def encode(self, *args):
        """
        1. 拼源串src，格式为 magic, *args，以split_bytes分隔。
        2. 使用aes_cbc方法将源串src加密成串dst
        3. 之后将iv与dst拼成token。
            token的格式为 前16位 固定为iv，剩下的部分为token。

        4. 之后将token进行base64加密方便http传输

        :param args:
        :return: 加密后字符串。为方便传输，类型为str。
        """
        iv = self._force_convert_to_bytes(self._gen_rand_str(self.block_size))

        aes = EasyAES(
            self.secret,
            iv
        )

        params = [self.magic]
        for it in args:
            params.append(self._force_convert_to_bytes(it))

        src = self.split_bytes.join(params)

        en_str = aes.encrypt(src)

        packed_str = iv + en_str

        base64_result = base64.b64encode(packed_str)

        return base64_result.decode('utf8')

    def decode(self, token):
        """
        解析token
        :param token:
        :return: 被加密的数组。与encode时一样的顺序，类型都是bytes。
        """

        packed_src = base64.b64decode(token)

        iv = packed_src[:self.block_size]
        en_src = packed_src[self.block_size:]

        aes = EasyAES(
            self.secret,
            iv
        )

        de_src = aes.decrypt(en_src)

        if not de_src:
            logger.info('decrypt fail. token: %s', token)
            return None

        values = de_src.split(self.split_bytes)

        # 第一个是magic
        magic = values.pop(0)
        if magic != self.magic:
            logger.info('invalid magic. token: %s', token)
            return None

        return values

    def _gen_rand_str(self, length):
        """
        generate secret key，参考django
        不区分大小写，因为mysql的unique不区分大小写
        """
        chars = 'abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*(-_=+)'
        return ''.join([random.choice(chars) for i in range(length)])

    def _force_convert_to_bytes(self, src):
        """
        强制转换为bytes
        bytes => bytes
        str => bytes
        else => str => bytes
        """

        if isinstance(src, bytes):
            return src
        elif isinstance(src, str):
            return bytes(src, encoding='utf8')
        else:
            return bytes(str(src), encoding='utf8')



"""
#=============================================================================
#
#     Filename: easy_aes.py
#      Summary: pip install pycrypto
#               使用 CBC 模式
#               测试: https://tools.lami.la/jiami/aes
#
#       Author: dantezhu
#        Email: dantezhu@qq.com
#     Homepage: http://www.vimer.cn
#
#      Created: 2019-01-08 21:29:21
#      Version: 0.0.1
#      History:
#               0.0.1 | dantezhu | 2019-01-08 21:29:21 | init
#
#=============================================================================
"""

from Crypto.Cipher import AES


class EasyAES:
    """封装AES，使用cbc模式"""

    _key = None
    _iv = None
    _mode = None

    def __init__(self, key, iv):
        """
        key 必须是bytes，长度必须为 128/192/256位，即16/24/32字节
        iv 必须是bytes，长度必须位 128位，即16字节
        """
        self._key = key
        self._iv = iv

    def encrypt(self, src):
        """
        src必须是bytes
        """
        # 填充字符
        padding = AES.block_size - len(src) % AES.block_size
        if padding > 0:
            # 填充
            src = src + padding * bytes(chr(padding), encoding='utf8')

        aes = AES.new(self._key, AES.MODE_CBC, self._iv)
        return aes.encrypt(src)

    def decrypt(self, src):
        """
        src必须是bytes
        """
        aes = AES.new(self._key, AES.MODE_CBC, self._iv)
        dst = aes.decrypt(src)
        if not dst:
            return b''

        if 0 < dst[-1] < AES.block_size:
            # 说明有填充
            return dst[0:-dst[-1]]
        else:
            return dst




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pansy.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pansy.proto',
  package='pansy.proto',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x0bpansy.proto\x12\x0bpansy.proto\"Y\n\nRspToUsers\x12)\n\x04rows\x18\x01 \x03(\x0b\x32\x1b.pansy.proto.RspToUsers.Row\x1a \n\x03Row\x12\x0c\n\x04uids\x18\x01 \x03(\x03\x12\x0b\n\x03\x62uf\x18\x02 \x01(\x0c\"\x1a\n\nCloseUsers\x12\x0c\n\x04uids\x18\x01 \x03(\x03\x62\x06proto3')
)




_RSPTOUSERS_ROW = _descriptor.Descriptor(
  name='Row',
  full_name='pansy.proto.RspToUsers.Row',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uids', full_name='pansy.proto.RspToUsers.Row.uids', index=0,
      number=1, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='buf', full_name='pansy.proto.RspToUsers.Row.buf', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=85,
  serialized_end=117,
)

_RSPTOUSERS = _descriptor.Descriptor(
  name='RspToUsers',
  full_name='pansy.proto.RspToUsers',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rows', full_name='pansy.proto.RspToUsers.rows', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_RSPTOUSERS_ROW, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28,
  serialized_end=117,
)


_CLOSEUSERS = _descriptor.Descriptor(
  name='CloseUsers',
  full_name='pansy.proto.CloseUsers',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uids', full_name='pansy.proto.CloseUsers.uids', index=0,
      number=1, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=119,
  serialized_end=145,
)

_RSPTOUSERS_ROW.containing_type = _RSPTOUSERS
_RSPTOUSERS.fields_by_name['rows'].message_type = _RSPTOUSERS_ROW
DESCRIPTOR.message_types_by_name['RspToUsers'] = _RSPTOUSERS
DESCRIPTOR.message_types_by_name['CloseUsers'] = _CLOSEUSERS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

RspToUsers = _reflection.GeneratedProtocolMessageType('RspToUsers', (_message.Message,), {

  'Row' : _reflection.GeneratedProtocolMessageType('Row', (_message.Message,), {
    'DESCRIPTOR' : _RSPTOUSERS_ROW,
    '__module__' : 'pansy_pb2'
    # @@protoc_insertion_point(class_scope:pansy.proto.RspToUsers.Row)
    })
  ,
  'DESCRIPTOR' : _RSPTOUSERS,
  '__module__' : 'pansy_pb2'
  # @@protoc_insertion_point(class_scope:pansy.proto.RspToUsers)
  })
_sym_db.RegisterMessage(RspToUsers)
_sym_db.RegisterMessage(RspToUsers.Row)

CloseUsers = _reflection.GeneratedProtocolMessageType('CloseUsers', (_message.Message,), {
  'DESCRIPTOR' : _CLOSEUSERS,
  '__module__' : 'pansy_pb2'
  # @@protoc_insertion_point(class_scope:pansy.proto.CloseUsers)
  })
_sym_db.RegisterMessage(CloseUsers)


# @@protoc_insertion_point(module_scope)



import os
import time
import queue
from pyglet.clock import Clock
import _thread

from ..share import constants
from ..share.log import logger
from .connection import Connection
from .request import Request
from ..share.task import Task
from ..proto.pansy_pb2 import RspToUsers, CloseUsers


class Worker:
    connection_class = Connection
    request_class = Request
    got_first_request = False

    task_queue = None

    station = None
    room_id = None

    conn = None
    # 业务可以通过clock来设置schedule
    clock = None

    # 工作进展
    work_progress = None

    def __init__(self, station, room_id):
        self.station = station
        self.room_id = room_id

        self.task_queue = queue.Queue(self.app.tasks_max_size)

        self.conn = self.connection_class(
            self,
            (self.app.host, self.app.port),
            self.app.conn_timeout
        )

        self.clock = Clock()

    def run(self):

        self.on_start()

        # daemon==True，主线程不需要等待网络线程结束
        _thread.start_new_thread(self.conn.run, ())
        _thread.start_new_thread(self._monitor_work_timeout, ())

        while self.enabled:
            try:
                self._handle()
            except KeyboardInterrupt:
                break
            except:
                logger.error('exc occur. worker: %s', self, exc_info=True)

        self.on_stop()

    def _handle(self):
        """
        主循环，等待网络消息，并进行处理
        """
        # 允许idle
        sleep_time = self.clock.get_sleep_time(True)
        # sleep_time is None 说明没有需要处理的，直接永久阻塞等待即可
        # 否则就要使用指定时间阻塞等待
        task_list = []
        # 第一个get要block
        block = True
        while True:
            try:
                task_list.append(
                    self.task_queue.get(block=block, timeout=sleep_time)
                )
            except queue.Empty:
                # 说明已经没有了
                break
            except KeyboardInterrupt:
                # 原样抛出
                raise
            except:
                # 说明出问题了
                logger.error('exc occur. worker: %s', self, exc_info=True)
                break
            finally:
                block = False

        for task in task_list:
            if task is None:
                # 通过传入None来中断block
                continue

            request = self.request_class(self, task)
            # 设置task开始处理的时间和信息
            self.work_progress = dict(
                begin_time=time.time(),
                request=request,
            )

            self._handle_request(request)

            self.work_progress = None

        self.clock.tick(True)

    def _handle_request(self, request):
        """
        出现任何异常的时候，服务器不再主动关闭连接
        """
        if not request.is_valid:
            return False

        if request.task.cmd == constants.CMD_CLIENT_CLOSED:
            self.app.events.close_client(request)
            for bp in self.app.blueprints:
                bp.events.close_app_client(request)
            return True

        if not request.view_func:
            logger.info('cmd invalid. request: %s' % request)
            return False

        if not self.got_first_request:
            self.got_first_request = True
            self.app.events.before_first_request(request)
            for bp in self.app.blueprints:
                bp.events.before_app_first_request(request)

        self.app.events.before_request(request)
        for bp in self.app.blueprints:
            bp.events.before_app_request(request)
        if request.blueprint:
            request.blueprint.events.before_request(request)

        if request.interrupted:
            # 业务要求中断
            return True

        view_func_exc = None

        try:
            request.view_func(request)
        except Exception as e:
            logger.error('view_func raise exception. e: %s, request: %s', e, request, exc_info=True)
            view_func_exc = e

        if request.blueprint:
            request.blueprint.events.after_request(request, view_func_exc)
        for bp in self.app.blueprints:
            bp.events.after_app_request(request, view_func_exc)
        self.app.events.after_request(request, view_func_exc)

        return True

    def _monitor_work_timeout(self):
        """
        监控task的耗时
        :return:
        """

        while True:
            time.sleep(1)

            work_progress = self.work_progress
            if work_progress:
                past_time = time.time() - work_progress['begin_time']
                if self.app.work_timeout is not None and past_time > self.app.work_timeout:
                    # 说明worker的处理时间已经太长了
                    logger.fatal('work timeout: %s / %s, request: %s',
                                 past_time, self.app.work_timeout, work_progress['request'])
                    # 不能这么做啊，直接把其他room也杀掉了，就打印一下告警算了
                    # 强制从子线程退出整个进程
                    # os._exit(-1)

    def push_task(self, task):
        """
        加入task，不阻塞
        return:
            True: 成功
            False: 失败
        """
        try:
            self.task_queue.put_nowait(task)
            return True
        except:
            logger.error('exc occur. worker: %s, task: %s', self, task, exc_info=True)
            return False

    @property
    def app(self):
        return self.station.app

    @property
    def enabled(self):
        return self.station.enabled

    def on_start(self):
        """
        当worker启动后
        可继承实现
        :return:
        """
        self.app.events.start_worker(self)
        for bp in self.app.blueprints:
            bp.events.start_app_worker(self)

    def on_stop(self):
        """
        当worker停止前
        可继承实现
        :return:
        """
        for bp in self.app.blueprints:
            bp.events.stop_app_worker(self)
        self.app.events.stop_worker(self)

    def write_to_users(self, data_list):
        """
        格式为
        [(uids, box), (uids, box), (uids, box) ...]
        :param data_list:
        :return:
        """

        msg = RspToUsers()

        for uids, data in data_list:
            if isinstance(data, self.app.box_class):
                data = data.pack()
            elif isinstance(data, dict):
                data = self.app.box_class(data).pack()

            row = msg.rows.add()
            row.buf = data
            row.uids.extend(uids)

        task = Task()
        task.cmd = constants.CMD_WRITE_TO_USERS
        task.body = msg.SerializeToString()

        return self.conn.write(task.pack())

    def close_users(self, uids):
        msg = CloseUsers()
        msg.uids.extend(uids)

        task = Task()
        task.cmd = constants.CMD_CLOSE_USERS
        task.body = msg.SerializeToString()

        return self.conn.write(task.pack())

    def get_fps(self):
        return self.clock.get_fps()

    def schedule(self, func, *args, **kwargs):
        self.clock.schedule(func, *args, **kwargs)

    def schedule_interval(self, func, interval, *args, **kwargs):
        self.clock.schedule_interval(func, interval, *args, **kwargs)

    def schedule_interval_soft(self, func, interval, *args, **kwargs):
        self.clock.schedule_interval_soft(func, interval, *args, **kwargs)

    def schedule_once(self, func, delay, *args, **kwargs):
        self.clock.schedule_once(func, delay, *args, **kwargs)

    def unschedule(self, func):
        self.clock.unschedule(func)

    def unschedule_all(self):
        self.clock = Clock()

    def __repr__(self):
        return '<%s name: %s, room_id: %r>' % (
            type(self).__name__, self.app.name, self.room_id
        )

